-----------------------------------------------------------------------
-- HEIG-VD, Haute Ecole d'Ingenerie et de Gestion du Canton de Vaud
-- Institut REDS
--
-- Composant        : bcd_adder
-- Description    : Additionneur BCD.
--                                Un paramètre générique permet de définir le nombre de digits.
--                                Un paramètre générique permet de forcer des erreurs
--                                du DUT.
--                                Il s'agit d'une implémentation structurelle exploitant
--                                le projet d'approfondissement de M. Fivaz.
-- Auteur             : Yann Thoma
-- Date                 : 01.03.2017
-- Version            : 1.0
--
-- Modification : -
--
-----------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use IEEE.math_real.all;
use work.bcd_pkg.all;
use work.ConfigurationPackage.all;

entity bcd_adder is
    generic (
        ERRNO : integer := 0;
    NDIGITS : integer := 4
    );
    port (
    input0_i : in bcd_number(NDIGITS - 1 downto 0);
    input1_i : in bcd_number(NDIGITS - 1 downto 0);
    result_o : out bcd_number(NDIGITS downto 0);
    hamming_o : out std_logic_vector(integer(ceil(log2(real(4*NDIGITS)))) downto 0)
    );
end bcd_adder;



architecture struct of bcd_adder is


    component GenericFullAdder is
        generic(
        Basis_g         : natural := 10 -- Number Basis_g 2^N_bits_c >= Basis_g
        );
        port(
        Carry_i             : in std_logic;
        Op_A_i            : in std_logic_vector(BCX_Bits(Basis_g)-1 downto 0);
        Op_B_i            : in std_logic_vector(BCX_Bits(Basis_g)-1 downto 0);
        Res_o         : out std_logic_vector(BCX_Bits(Basis_g)-1 downto 0);
        Carry_o             : out std_logic
        );

    end component;

    signal carry_s : std_logic_vector(NDIGITS downto 0);

    signal result_s : bcd_number(NDIGITS downto 0);

    begin

        process(input0_i, input1_i) is
            variable sum_v : integer;
            begin
                sum_v := 0;
                for digit in 0 to NDIGITS - 1 loop
                    for abit in 0 to 3 loop
                        if (input0_i(digit)(abit) /= input1_i(digit)(abit)) then
                            sum_v := sum_v + 1;
                        end if;
                    end loop;
                end loop;

                if (ERRNO = 3) then
                    if (unsigned(input0_i(0)) > 9) then
                        sum_v := sum_v - 1;
                    end if;
                end if;

                if (ERRNO = 4) then
                        sum_v := sum_v - 1;
                end if;

                hamming_o <= std_logic_vector(to_unsigned(sum_v, hamming_o'length));
            end process;

            carry_s(0) <= '0';

            result_s(NDIGITS)(0) <= carry_s(NDIGITS) when ERRNO /= 5 else not carry_s(NDIGITS);
            result_s(NDIGITS)(3 downto 1) <= (others => '0');

            process(result_s) is
                begin
                    for digit in 0 to NDIGITS loop
                        result_o <= result_s;
                    end loop;

                    if (ERRNO = 1) then
                        if (unsigned(input0_i(0)) > 9) then
                            result_o(NDIGITS - 1)(1) <= '0';
                        end if;
                    end if;

                    if (ERRNO = 2) then
                        if (unsigned(input0_i(0)) > 9) then
                            result_o(0)(2) <= 'X';
                        end if;
                    end if;

                    if (ERRNO = 6) then
                        result_o(0)(2) <= not result_s(0)(2);
                    end if;

                    if (ERRNO = 7) then
                        result_o(NDIGITS - 1)(2) <= not result_s(NDIGITS - 1)(2);
                    end if;

                end process;

                gen_adders : for digit in 0 to NDIGITS -1 generate

                    adder : GenericFullAdder
                    generic map (Basis_g => 10)
                    port map (
                    carry_i => carry_s(digit),
                    Op_A_i => input0_i(digit),
                    Op_B_i => input1_i(digit),
                    Res_o => result_s(digit),
                    Carry_o => carry_s(digit + 1)
                    );
                end generate;

            end struct;
